## ---- include = FALSE---------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    eval = TRUE,
    echo = TRUE,
    fig.width = 7,
    fig.height = 5,
    fig.align = 'center'
)

## ---- eval = FALSE------------------------------------------------------------
#  # install.packages('remotes')
#  remotes::install_github("muschellij2/NiftiArray")

## ---- warning = FALSE, message = FALSE----------------------------------------
library(DelayedArray)
library(DelayedMatrixStats)
library(dplyr)
library(forcats)
library(ggplot2)
library(gridExtra)
library(kableExtra)
library(knitr)
library(microbenchmark)
library(RNifti)
library(NiftiArray)
library(neurobase)
library(pbapply)
library(profmem)
library(remotes)
library(stringr)
library(tibble)
library(tidyr)
nifti_header = NiftiArray::nifti_header

## -----------------------------------------------------------------------------
# Information about URL to download and where to save the image locally in destiation
urls = file.path("https://raw.githubusercontent.com", 
                     "muschellij2", "open_ms_data", "master", 
                     "cross_sectional", "MNI", paste0("patient0", 1:5),
                     "FLAIR_N4_noneck_reduced_winsor_regtoFLAIR_brain_N4_regtoMNI.nii.gz")
nii_destinations = sapply(urls, function(x) tempfile(fileext = ".nii.gz"))
hdf5_destinations = sub(".nii.gz", ".h5", nii_destinations)

fileinfo = tibble::tibble(url = urls, 
                          nii_destination = nii_destinations,
                          hdf5_destination = hdf5_destinations)
# Download all the files to the nii_destination
mapply(function(x,y) {
    download.file(url = x, destfile = y)
}, fileinfo$url, fileinfo$nii_destination)

# Notice the files were saved to a temporary directory on your machine
# Change tempdir() to the directory where you saved the files if you adapted nii_destination
list.files(tempdir(), full.names = TRUE)

## ---- echo = FALSE------------------------------------------------------------
if (Sys.info()[["user"]] == 'alval') {
    # Information about URL to download and where to save the image locally in destiation
    data_dir = '/Users/alval/Box/rpackages/NiftiArray_examples/data/'
    fileinfo = tibble::tibble(
        url = fileinfo$url,
        nii_destination = paste0(data_dir, "nii/", 
                                 "flairpatient0", 1:5, ".nii.gz"),
        hdf5_destination = paste0(data_dir, "hdf5/", 
                                  "flairpatient0", 1:5, ".h5"))
}
temp = neurobase::readnii(fileinfo$nii_destination[1])

## -----------------------------------------------------------------------------
# Write the NiftiArray object to disk and load it into R as patient01
patient01 = NiftiArray::writeNiftiArray(
    fileinfo$nii_destination[1], 
    filepath = fileinfo$hdf5_destination[1], 
    overwrite = TRUE)
file.exists(fileinfo$hdf5_destination[1])

## -----------------------------------------------------------------------------
class(patient01)

## -----------------------------------------------------------------------------
# Print the NIfTI header associated with patient01
NiftiArray::nifti_header(patient01)

## -----------------------------------------------------------------------------
# Remove patient01 object from R memory so it can be re-loaded later
rm(patient01)

## -----------------------------------------------------------------------------
# Reload patient01 using NiftiArray::NiftiArray from a NIfTI object on disk
patient01 = NiftiArray::NiftiArray(fileinfo$nii_destination[1])
# Remove patient01 object from R memory so it can be re-loaded later
rm(patient01)

## -----------------------------------------------------------------------------
# Reload patient01 using NiftiArray::NiftiArray from a HDF5 object on disk
patient01 = NiftiArray::NiftiArray(fileinfo$hdf5_destination[1])

## -----------------------------------------------------------------------------
# TODO explain subsetting and loading into memory
# drop = FALSE - https://github.com/Bioconductor/DelayedArray/issues/6
# Index a single voxel
patient01[90, 89, 101]
# Change the value of the background voxels from 0 to -100
## Note: Normally this should be done using a brain mask but this is just an example for this tutorial
patient01[patient01 == 0] = -100
# Summary stats
min(patient01)
max(patient01)
# Image operations
patient_sum = patient01 + patient01
# Notice it did sum the NiftiArray objects voxel-wise across objects
patient_sum[90, 89, 101]

## -----------------------------------------------------------------------------
# Due to the block calculation some base functions will not be available
# The table function errors
# TODO: Does DelayedArray have a table? DelayedArray::table()
# table(patient01)
table(c(patient01 > 5))

## -----------------------------------------------------------------------------
# Remove patient01 object from R memory so it can be re-loaded later
rm(patient01)

## ----listmap------------------------------------------------------------------
# Write out the entire filepath of NIfTI objects to HDF5s
all_patients_list = mapply(function(x,y) {
    writeNiftiArray(x = x, filepath = y, overwrite = TRUE)
}, fileinfo$nii_destination, fileinfo$hdf5_destination)
# List the files in the temporary directory -- all subjects should be there as NIfTIs and .h5s
list.files(tempdir(), pattern = ".(nii|h5)")
# Remove all_patients_list since we will load it in a different way later
rm(all_patients_list)

## ----nilist2------------------------------------------------------------------
# Convert, write to temporary disk, and load all_patients_list using NiftiArray::NiftiArrayList
all_patients_list = NiftiArray::NiftiArrayList(fileinfo$nii_destination)
# Show the class is NiftiArrayList
class(all_patients_list)

## ----niArrayList--------------------------------------------------------------
# Load all_patients_list using NiftiArray::NiftiArrayList
all_patients_list = NiftiArray::NiftiArrayList(fileinfo$hdf5_destination)
# Show the class is NiftiArrayList
class(all_patients_list)

## ----niarray------------------------------------------------------------------
# Reload patient01 using NiftiArray::NiftiArray from a HDF5 object on disk
patient01 = NiftiArray::NiftiArray(fileinfo$hdf5_destination[[1]])
# Convert the NiftiArray to a NiftiMatrix
patient01_niimat = as(patient01, "NiftiMatrix")
# Verify patient01_niimat is of class NiftiMatrix
class(patient01_niimat)
# Notice the NiftiMatrix has only 1 column
dim(patient01_niimat)
# Index the NiftiMatrix like a vector
patient01_niimat[40000:40010]
# The NiftiMatrix object is still memory efficient
object.size(patient01_niimat)
# Obtain the NIfTI header
NiftiArray::nifti_header(patient01_niimat)

## ----acbind-------------------------------------------------------------------
# Show all_patients_list is of NiftiArrayList
class(all_patients_list)
# Convert each element in the list to a NiftiMatrix
all_patients_niimat = pbapply::pblapply(all_patients_list, as, "NiftiMatrix")
# Concatenate the list elements column-wise
all_patients_niimat  = do.call(DelayedArray::acbind, all_patients_niimat)
# Show the matrix
head(all_patients_niimat)
# Notice the dimension is all the voxels in the brain by the 5 patients we are working with
dim(all_patients_niimat)
# Obtain the NIfTI header
NiftiArray::nifti_header(all_patients_niimat)

## ----vmedians-----------------------------------------------------------------
all_patients_niimat = as(all_patients_niimat, "HDF5Matrix")
# Calculate the voxel-wise  median across subjects
voxel_medians = DelayedMatrixStats::rowMedians(all_patients_niimat)
# Index the median vector
voxel_medians[1043001]
# Show the class of voxel_medians
class(voxel_medians)
# Calculate the object size
object.size(voxel_medians)

## -----------------------------------------------------------------------------
# Convert the median vector back to a `NiftiMatrix`
voxel_medians = as(voxel_medians, 'NiftiMatrix')
# Notice the size is back to a memory efficient value
object.size(voxel_medians)
# The header is no longer accurate though
NiftiArray::nifti_header(voxel_medians)

## ----vmeans-------------------------------------------------------------------
# Calculate the voxel-wise  mean across subjects
voxel_means = DelayedMatrixStats::rowMeans2(all_patients_niimat)

## ----as_to_array--------------------------------------------------------------
class(patient01_niimat)
niimat2niiarr = as(patient01_niimat, "NiftiArray")
class(niimat2niiarr)

## ----reshaping_again----------------------------------------------------------
class(patient01_niimat)
niimat2nii = as(patient01_niimat, "niftiImage")
class(niimat2nii)

## ----niMatrix-----------------------------------------------------------------
class(patient01)
niiarr2niimat = as(patient01, "NiftiMatrix")
class(niiarr2niimat)

## ----niImage------------------------------------------------------------------
class(patient01)
niiarr2nii = as(patient01, "niftiImage")
class(niiarr2nii)

## ----hdr----------------------------------------------------------------------
# Median
# Obtain the image header associated with the all_patients_list NiftiArrayList
image_header = NiftiArray::nifti_header(all_patients_list)
# Initialize an array
median_arr = array(voxel_medians, dim = image_header$dim[2:4])
# The object size of median_arr is normal because it is a normal array
object.size(median_arr)
# Convert the array to a NiftiArray and write it out to a temporary file
median_niiarr = NiftiArray::writeNiftiArray(median_arr, header = image_header)
# The object size is back to our efficient NiftiArray
object.size(median_niiarr)

# Mean
# Obtain the image header associated with the all_patients_list NiftiArrayList
image_header = NiftiArray::nifti_header(all_patients_list)
# Initialize an array
mean_arr = array(voxel_means, dim = image_header$dim[2:4])
# The object size of mean_arr is normal because it is a normal array
object.size(mean_arr)
# Convert the array to a NiftiArray and write it out to a temporary file
mean_niiarr = NiftiArray::writeNiftiArray(mean_arr, header = image_header)
# The object size is back to our efficient NiftiArray
object.size(mean_niiarr)

## ----medianimg----------------------------------------------------------------
median_nii = as(median_niiarr, "niftiImage")
mean_nii = as(mean_niiarr, "niftiImage")

## -----------------------------------------------------------------------------
neurobase::ortho2(median_nii, pdim = nifti_header(median_nii)$pixdim)
neurobase::ortho2(mean_nii, pdim = nifti_header(mean_nii)$pixdim)

## ----objsize, echo = TRUE-----------------------------------------------------
# Calculate the a single patient object size using different NIfTI read functions
memory = tibble::tibble(read_type = c('NiftiArray::NiftiArray', 
                                      'NiftiArray::writeNiftiArray', 
                                      'RNifti::readNifti', 
                                      'neurobase::readnii'),
                        byte_size = c(object.size(NiftiArray::NiftiArray(fileinfo$nii_destination[1])),
                                      object.size(NiftiArray::writeNiftiArray(fileinfo$nii_destination[1])),
                                      object.size(RNifti::readNifti(fileinfo$nii_destination[1])),
                                      object.size(neurobase::readnii(fileinfo$nii_destination[1]))),
                       log_byte_size = log10(byte_size)) %>% 
  dplyr::mutate(read_type = as.factor(read_type),
                read_type = forcats::fct_reorder(read_type, byte_size, .desc = TRUE))

# Table of memory
knitr::kable(
  memory,
  col.names = c('Read Function', 'Byte Size', 'log_10(Byte Size)'),
  format = 'html',
  digits = 2, 
  caption = 'Memory mapped in bytes from a single patients image read. 
  The image dimension is 182 by 218 by 182 with pixel dimension 1 mm by 1 mm by 1 mm. 
  On disk the image is 4.9 MB.', 
  booktabs = TRUE
) %>% 
  kableExtra::kable_styling("striped", full_width = FALSE) 

## -----------------------------------------------------------------------------
# Bar graph in bytes
ggplot(memory, aes(x = read_type, y = byte_size, fill = read_type)) +
  geom_bar(stat="identity") +
  geom_text(aes(label=byte_size), vjust=-0.3, size=3.5) +
  ylab('Memory in Bytes') +
  xlab('Read Function') +
  labs(title = 'Memory Comparison for Reading a Single NIfTI Image',
       subtitle = 'Image Dimension: 182 by 218 by 182, Pixel Dimension: 1 mm by 1 mm by 1 mm') +
  theme_minimal() +
  theme(plot.title = element_text(hjust=0.5), legend.position = 'none')

# Bar graph in log(bytes)
ggplot(memory, aes(x = read_type, y = log_byte_size, fill = read_type)) +
  geom_bar(stat="identity") +
  geom_text(aes(label=round(log_byte_size), digits = 2), vjust=-0.3, size=3.5) +
  ylab('Memory in log(Bytes)') +
  xlab('Read Function') +
  labs(title = 'Memory Comparison for Reading a Single NIfTI Image on the Log Scale',
       subtitle = 'Image Dimension: 182 by 218 by 182, Pixel Dimension: 1 mm by 1 mm by 1 mm') +
  theme_minimal() +
  theme(plot.title = element_text(hjust=0.5), legend.position = 'none')

## ----profmem------------------------------------------------------------------
# Initalize a list to store the memory profile
memory_profile = list()
# Memory profile of RNifti::readNifti()
memory_profile[[1]] = tibble::tibble(read_type = 'RNifti::readNifti', 
                                     memory_profile = profmem::profmem(RNifti::readNifti(fileinfo$nii_destination[1]))$bytes)
memory_profile[[2]] = tibble::tibble(read_type = 'neurobase::readnii', 
                                     memory_profile = profmem::profmem(neurobase::readnii(fileinfo$nii_destination[1]))$bytes)
memory_profile[[3]] = tibble::tibble(read_type = 'NiftiArray::NiftiArray - NIfTI File', 
                                     memory_profile = profmem::profmem(NiftiArray::NiftiArray(fileinfo$nii_destination[1]))$bytes)
memory_profile[[4]] = tibble::tibble(read_type = 'NiftiArray::writeNiftiArray', 
                                     memory_profile = profmem::profmem(NiftiArray::writeNiftiArray(fileinfo$nii_destination[1]))$bytes)
memory_profile[[5]] = tibble::tibble(read_type = 'NiftiArray::NiftiArray - HDF5 File',
                                     memory_profile = profmem::profmem(NiftiArray::NiftiArray(fileinfo$hdf5_destination[1]))$bytes)

memory_profile = dplyr::bind_rows(memory_profile) %>% 
  tidyr::drop_na() %>% 
  dplyr::group_by(read_type) %>% 
  dplyr::mutate(x = dplyr::row_number())

## ----plotmem------------------------------------------------------------------
ggplot(data=memory_profile, aes(x=x, y=log10(memory_profile))) +
  geom_line()+
  geom_point() +
  facet_wrap(~read_type, scales = "free_x")
ggplot(data=memory_profile, aes(x=x, y=log10(memory_profile))) +
  geom_line()+
  geom_point() +
  facet_wrap(~read_type, scales = "free")

## -----------------------------------------------------------------------------
speed = tibble::as_tibble(
  microbenchmark::microbenchmark(
    NiftiArray::NiftiArray(fileinfo$hdf5_destination[1]),
    NiftiArray::NiftiArray(fileinfo$nii_destination[1]),
    NiftiArray::writeNiftiArray(fileinfo$nii_destination[1]),
    RNifti::readNifti(fileinfo$nii_destination[1]),
    neurobase::readnii(fileinfo$nii_destination[1]),
    times = 5)
  ) %>% 
  dplyr::mutate(time = time/(1*10^9)) %>% # conver nanoseconds to seconds
  dplyr::rename(read_type = expr) %>% 
  dplyr::mutate(read_type = 
                  dplyr::case_when(
                    stringr::str_detect(read_type, 'RNifti') ~ "RNifti::readNifti",
                    stringr::str_detect(read_type, 'neurobase') ~ "neurobase::readnii",
                    stringr::str_detect(read_type, 'hdf5') ~ "NiftiArray::NiftiArray - HDF5 File",                    
                    stringr::str_detect(read_type, 'NiftiArray::NiftiArray\\(fileinfo\\$nii') ~ "NiftiArray::NiftiArray - NIfTI File",
                    stringr::str_detect(read_type, 'NiftiArray::writeNiftiArray') ~ "NiftiArray::writeNiftiArray"),
                read_type = as.factor(read_type),
                read_type = forcats::fct_reorder(read_type, time, .desc = TRUE))

# Table of memory
speed_summary = speed %>% 
  dplyr::group_by(read_type) %>% 
  dplyr::summarise(
    mean = mean(time, na.rm = TRUE),
    median = mean(time, na.rm = TRUE),
    sd = sd(time, na.rm = TRUE),
    min = min(time, na.rm = TRUE),
    max = max(time, na.rm = TRUE)
  )

knitr::kable(
  speed_summary,
  col.names = c('Read Function', 'Mean', 'Median', 'Std. Dev.', 'Min.', 'Max.'),
  format = 'html',
  digits = 2,
  caption = 'Memory mapped in bytes from a single patients image read.
  The image dimension is 182 by 218 by 182 with pixel dimension 1 mm by 1 mm by 1 mm.
  On disk the image is 4.9 MB.',
  booktabs = TRUE
) %>%
  kableExtra::kable_styling("striped", full_width = FALSE)

## -----------------------------------------------------------------------------
ggplot(data = speed, aes(x = read_type, y = time)) +
  geom_boxplot() +   
  coord_flip() +
  ylab('Time (Seconds)') +
  xlab('Read Function') +
  labs(title = 'Speed Comparison for Reading a Single NIfTI Image',
       subtitle = 'Image Dimension: 182 by 218 by 182, Pixel Dimension: 1 mm by 1 mm by 1 mm') +
  theme_minimal() +
  theme(plot.title = element_text(hjust=0.5), legend.position = 'none')

